#pragma once

namespace c4u {

#define MIN_SHELL_ID 1
#define MAX_SHELL_ID 0xffffffff

class context
{
public:
	context () : m_pMalloc (NULL)
	{
		SHGetMalloc (&m_pMalloc);
	}
	~context ()
	{
		if (m_pMalloc)
			m_pMalloc->Release ();
	}

	inline bool DoExplorerMenu (HWND hwnd, int csidl, POINT point)
	{
		LPITEMIDLIST pidlMain = NULL;
		if (SUCCEEDED (SHGetSpecialFolderLocation (NULL, csidl, &pidlMain)) &&
			pidlMain) {
			return DoExplorerMenu (hwnd, pidlMain, point);
		}
		return false;
	}

	inline bool DoExplorerMenu (HWND hwnd, LPCTSTR pszPath, POINT point)
	{
		TCHAR tchPath [MAX_PATH];
		GetFullPathName (pszPath, sizeof (tchPath) / sizeof (TCHAR), tchPath, NULL);
		bool bResult = false;
		LPSHELLFOLDER psfFolder = NULL;
		if (SUCCEEDED (SHGetDesktopFolder (&psfFolder))) {

			// Convert the path name into a pointer to an item ID list (pidl).
			LPITEMIDLIST pidlMain = NULL;
			ULONG ulCount, ulAttr;
			if (SUCCEEDED (psfFolder->ParseDisplayName (NULL,
				NULL, CT2W (tchPath), &ulCount, &pidlMain, &ulAttr)) && pidlMain) {

				bResult = DoExplorerMenu (hwnd, pidlMain, point);
			}
			psfFolder->Release ();
		}
		return bResult;
	}

protected:
	// global variables used for passing data to the subclassing wndProc
	WNDPROC g_pOldWndProc;			// regular window proc
	LPCONTEXTMENU2 g_pIContext2;	// active shell context menu
	LPMALLOC m_pMalloc;

	bool DoExplorerMenu (HWND hwnd, LPITEMIDLIST pidlMain, POINT point)
	{
		bool bResult = false;
		
		LPSHELLFOLDER psfFolder = NULL;
		if (FAILED (SHGetDesktopFolder (&psfFolder))) {
			return false;
		}

		UINT nCount = GetItemCount (pidlMain);
		if (nCount) {

			// Initialize psfFolder with a pointer to the IShellFolder
			// interface of the folder that contains the item whose context
			// menu we're after, and initialize pidlItem with a pointer to
			// the item's item ID. If nCount > 1, this requires us to walk
			// the list of item IDs stored in pidlMain and bind to each
			// subfolder referenced in the list.
			LPITEMIDLIST pidlItem = pidlMain;
			while (--nCount) {						

				// Create a 1-item item ID list for the next item in pidlMain.
				LPITEMIDLIST pidlNextItem = DuplicateItem (m_pMalloc, pidlItem);
				if (!pidlNextItem) {
					m_pMalloc->Free (pidlMain);
					psfFolder->Release ();
					return bResult;
				}

				// Bind to the folder specified in the new item ID list.
				LPSHELLFOLDER psfNextFolder = NULL;
				if (FAILED (psfFolder->BindToObject (pidlNextItem,
					NULL, IID_IShellFolder, (void**) &psfNextFolder))) {
					m_pMalloc->Free (pidlNextItem);
					m_pMalloc->Free (pidlMain);
					psfFolder->Release ();
					return bResult;
				}

				// Release the IShellFolder pointer to the parent folder
				// and set psfFolder equal to the IShellFolder pointer for
				// the current folder.
				psfFolder->Release ();
				psfFolder = psfNextFolder;

				// Release the storage for the 1-item item ID list we created
				// just a moment ago and initialize pidlItem so that it points
				// to the next item in pidlMain.
				m_pMalloc->Free (pidlNextItem);
				pidlItem = GetNextItem (pidlItem);
			}

			if (!nCount) {
				// Get a pointer to the item's IContextMenu interface and call
				// IContextMenu::QueryContextMenu to initialize a context menu.
				int cmType;
				LPCONTEXTMENU pContextMenu = NULL;
				if (SUCCEEDED (GetSHContextMenu (psfFolder, pidlItem,
					(void**) &pContextMenu, &cmType))) {

					HMENU hMenu = CreatePopupMenu ();
					if (hMenu) {
						if (SUCCEEDED (pContextMenu->QueryContextMenu (hMenu, 0,
							MIN_SHELL_ID, MAX_SHELL_ID, CMF_NORMAL))) {

							// install the subclassing "hook", for versions 2
							// or 3
							if (cmType > 1) {
								::SetWindowLongPtr (hwnd, GWLP_USERDATA,
									(LONG_PTR) this);
								g_pIContext2 = (LPCONTEXTMENU2) pContextMenu;
								g_pOldWndProc = (WNDPROC) ::SetWindowLongPtr (hwnd,
									GWLP_WNDPROC, (LONG_PTR) HookWndProc);
							} else
								g_pOldWndProc = NULL;

							ClientToScreen (hwnd, &point);
							UINT nCmd = TrackPopupMenu (hMenu, TPM_LEFTALIGN |
								TPM_LEFTBUTTON | TPM_RIGHTBUTTON | TPM_RETURNCMD,
								point.x, point.y, 0, hwnd, NULL);

							// restore old wndProc
							if (g_pOldWndProc)
								::SetWindowLongPtr (hwnd, GWLP_WNDPROC,
									(LONG_PTR) g_pOldWndProc);

							// If a command was selected from the menu,
							// execute it.
							if (nCmd) {
								CMINVOKECOMMANDINFO ici;
								ZeroMemory (&ici, sizeof (ici));
								ici.cbSize = sizeof (CMINVOKECOMMANDINFO);
								ici.hwnd = hwnd;
								ici.lpVerb = (LPCSTR) MAKEINTRESOURCE (nCmd - 1);
								ici.nShow = SW_SHOWNORMAL;
								bResult = SUCCEEDED (
									pContextMenu->InvokeCommand (&ici));
							}
						}
						DestroyMenu (hMenu);
					}
					g_pIContext2 = NULL;
					pContextMenu->Release ();
				} // if (SUCCEEDED (GetSHContextMenu
			} // if (!nCount) {
		} // if (nCount) {
		psfFolder->Release ();
		return bResult;
	}

	inline LPITEMIDLIST GetNextItem (LPITEMIDLIST pidl)
	{
		USHORT nLen;
		if ((nLen = pidl->mkid.cb) == 0)
			return NULL;	    
		return (LPITEMIDLIST) (((LPBYTE) pidl) + nLen);
	}

	inline UINT GetItemCount (LPITEMIDLIST pidl)
	{
		USHORT nLen;
		UINT nCount = 0;
		while ((nLen = pidl->mkid.cb) != 0) {
			pidl = GetNextItem (pidl);
			nCount++;
		}
		return nCount;
	}

	inline LPITEMIDLIST DuplicateItem (LPMALLOC pMalloc, LPITEMIDLIST pidl)
	{
		USHORT nLen = pidl->mkid.cb;
		if (!nLen)
			return NULL;

		LPITEMIDLIST pidlNew = (LPITEMIDLIST) pMalloc->Alloc (nLen + sizeof (USHORT));
		if (!pidlNew)
			return NULL;

		CopyMemory (pidlNew, pidl, nLen);
		*((USHORT*) (((LPBYTE) pidlNew) + nLen)) = 0;
		return pidlNew;
	}

	inline HRESULT GetSHContextMenu (LPSHELLFOLDER psfFolder,
		LPCITEMIDLIST localPidl, void** ppCM, int* pcmType)
	{
		*ppCM = NULL;
		LPCONTEXTMENU pICv1 = NULL; // plain version
		// try to obtain the lowest possible IContextMenu
		HRESULT hr = psfFolder->GetUIObjectOf (NULL, 1, &localPidl, 
			IID_IContextMenu, NULL, (void**)&pICv1);
		if(pICv1) {
			// try to obtain a higher level pointer, first 3 then 2
			hr = pICv1->QueryInterface (IID_IContextMenu3, ppCM);
			if (NOERROR == hr)
				*pcmType = 3;
			else {
				hr = pICv1->QueryInterface (IID_IContextMenu2, ppCM);
				if(NOERROR == hr)
					*pcmType = 2;
			}
			if(*ppCM)
				pICv1->Release(); // free initial "v1.0" interface
			else {
				// no higher version supported
				*pcmType = 1;
				*ppCM = pICv1;
				hr = NOERROR; // never mind the query failures, this'll do
			}
		}
		return hr;
	}

	static LRESULT CALLBACK HookWndProc (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
	{
		context* self = (context*) ::GetWindowLongPtr (hwnd, GWLP_USERDATA);	
		ASSERT (self);
		switch (msg) { 
			case WM_DRAWITEM:
			case WM_MEASUREITEM:
				if (wp)
					break;
			case WM_INITMENUPOPUP:
				self->g_pIContext2->HandleMenuMsg (msg, wp, lp);
				return (msg == WM_INITMENUPOPUP ? 0 : TRUE);
		}
		return ::CallWindowProc (self->g_pOldWndProc, hwnd, msg, wp, lp);
	}
};

}
